--
-- Pixel art rendering and camera movement example.
-- Shaders could be used but aren't in this example.
--

local CAMERA_WIDTH  = 256 -- Based on superMetroid.png
local CAMERA_HEIGHT = 222

local PLAYER_MOVE_SPEED   = 60 -- Pixels per second.
local CAMERA_FOLLOW_SPEED = 5



local canvas, canvasFilter, canvasSubpixels
local world, player, camera

local integerScaling = false



local function createCanvas(filter, subpixels)
	canvasFilter    = filter
	canvasSubpixels = subpixels

	canvas = love.graphics.newCanvas(CAMERA_WIDTH*canvasSubpixels, CAMERA_HEIGHT*canvasSubpixels)
	canvas:setFilter(canvasFilter, canvasFilter)
end

function love.load()
	love.graphics.setDefaultFilter("nearest", "nearest")

	-- The origin point of objects is in the bottom center.
	world  = {x=CAMERA_WIDTH/2, y=CAMERA_HEIGHT  , sprite=love.graphics.newImage("superMetroid.png")}
	player = {x=CAMERA_WIDTH/2, y=CAMERA_HEIGHT/2, sprite=love.graphics.newImage("megaman.png")}

	-- The camera's position represents the center of the view (so if the
	-- camera and the player share the same position then the player is
	-- centered on the screen).
	camera = {x=player.x, y=player.y}

	createCanvas("nearest", 1)
end



function love.keypressed(key)
	if key == "1" then
		-- Sharpest pixels, but pixels will likely be different sizes.
		createCanvas("nearest", 1)

	elseif key == "2" then
		-- Linear filter with no subpixels will usually result in ugly blur.
		createCanvas("linear", 1)

	elseif key == "3" then
		-- Use subpixels with linear filter for a little anti-aliasing so
		-- pixel sizes look more uniform, but everything is still pretty sharp.
		createCanvas("linear", 2)

	elseif key == "4" then
		-- Larger difference between camera and window size require more
		-- subpixels to prevent blurriness. (The subpixel amount could be
		-- calculated dynamically from the window height.)
		createCanvas("linear", 4)

	elseif key == "i" then
		integerScaling = not integerScaling
		-- Note: Subpixels are not needed if integer scaling is enabled.

	elseif key == "escape" then
		love.event.quit()
	end
end



local function lerp(v1, v2, t)
	return v1 + (v2-v1) * t
end
local function damp(current, target, lambda, dt)
	return lerp(current, target, 1-math.exp(-lambda*dt)) -- https://www.rorydriscoll.com/2016/03/07/frame-rate-independent-damping-using-lerp/
end

function love.update(dt)
	local moveDirX = 0
	local moveDirY = 0

	if love.keyboard.isDown"left"  then  moveDirX = moveDirX - 1  end
	if love.keyboard.isDown"right" then  moveDirX = moveDirX + 1  end
	if love.keyboard.isDown"up"    then  moveDirY = moveDirY - 1  end
	if love.keyboard.isDown"down"  then  moveDirY = moveDirY + 1  end

	player.x = player.x + PLAYER_MOVE_SPEED * moveDirX * dt
	player.y = player.y + PLAYER_MOVE_SPEED * moveDirY * dt

	camera.x = damp(camera.x, player.x, CAMERA_FOLLOW_SPEED, dt) -- Move towards player smoothly.
	camera.y = damp(camera.y, player.y, CAMERA_FOLLOW_SPEED, dt) -- Do camera.x=player.x etc. for instant snap.
end



local function getCanvasRenderInfo()
	local windowWidth, windowHeight = love.graphics.getDimensions()
	local canvasWidth, canvasHeight = canvas:getDimensions()

	-- Fill as much of the window as possible with the canvas while preserving the aspect ratio.
	local canvasScale = math.min(windowWidth/canvasWidth, windowHeight/canvasHeight)
	-- local canvasScale = windowHeight/canvasHeight -- This would fill the height and possibly cut off the sides.

	if integerScaling then
		canvasScale = math.floor(canvasScale * canvasSubpixels) / canvasSubpixels
		canvasScale = math.max(canvasScale, 1/canvasSubpixels) -- Avoid canvasScale=0 if the window is tiny!
	end

	local canvasWidthScaled  = canvasWidth  * canvasScale
	local canvasHeightScaled = canvasHeight * canvasScale

	local canvasX = math.floor((windowWidth  - canvasWidthScaled ) / 2) -- Center the canvas on the screen.
	local canvasY = math.floor((windowHeight - canvasHeightScaled) / 2)

	local pixelScale = canvasScale * canvasSubpixels -- Scale of big pixels, not sub-pixels.

	return canvasX, canvasY, canvasScale, pixelScale -- These values could be cached until a setting or the window size changes.
end



local function screenToWorld(x, y)
	local canvasX, canvasY, canvasScale, pixelScale = getCanvasRenderInfo()

	local cornerX = camera.x - CAMERA_WIDTH/2 -- Top left corner of in-world camera.
	local cornerY = camera.y - CAMERA_HEIGHT/2

	x = ((x - canvasX) / pixelScale) + cornerX
	y = ((y - canvasY) / pixelScale) + cornerY

	return x, y
end

local function worldToScreen(x, y)
	local canvasX, canvasY, canvasScale, pixelScale = getCanvasRenderInfo()

	local cornerX = camera.x - CAMERA_WIDTH/2 -- Top left corner of in-world camera.
	local cornerY = camera.y - CAMERA_HEIGHT/2

	x = ((x - cornerX) * pixelScale) + canvasX -- Inverse of screenToWorld.
	y = ((y - cornerY) * pixelScale) + canvasY

	return x, y
end



local function roundCoords(x, y)
	-- Round to nearest subpixel.
	return math.floor(x * canvasSubpixels + .5) / canvasSubpixels,
	       math.floor(y * canvasSubpixels + .5) / canvasSubpixels
end

local function drawEntity(entity)
	local x, y    = roundCoords(entity.x, entity.y)
	local originX = entity.sprite:getWidth() / 2
	local originY = entity.sprite:getHeight()

	love.graphics.draw(entity.sprite, x,y, 0, 1,1, originX,originY)
	love.graphics.circle("fill", x,y, 5) -- Show where the sprite's origin is.
end

function love.draw()
	--
	-- Render game world.
	--
	love.graphics.push("all")
	love.graphics.setCanvas(canvas)
	love.graphics.clear(0, .15, 0, 1)
	love.graphics.scale(canvasSubpixels)

	-- Scroll to the camera's position.
	local cornerX = camera.x - CAMERA_WIDTH/2
	local cornerY = camera.y - CAMERA_HEIGHT/2
	love.graphics.translate(roundCoords(-cornerX, -cornerY)) -- Negative values because as the camera moves to the right, the world moves to the left etc.

	-- Render sprites/entities.
	drawEntity(world)
	drawEntity(player)

	-- Show where the center of the world is.
	-- (Should be seen in the top left corner initially.)
	love.graphics.setColor(1, 0, 0)
	love.graphics.circle("fill", 0, 0, 5)

	love.graphics.pop()

	--
	-- Render canvas.
	--
	local canvasX, canvasY, canvasScale, pixelScale = getCanvasRenderInfo()

	love.graphics.clear(.2, .2, .2, 1)
	love.graphics.draw(canvas, canvasX, canvasY, 0, canvasScale)

	-- Info.
	local playerXOnScreen, playerYOnScreen = worldToScreen(player.x, player.y)
	local cameraXOnScreen, cameraYOnScreen = worldToScreen(camera.x, camera.y)
	local mouseXOnScreen , mouseYOnScreen  = love.mouse.getPosition()
	local mouseXInWorld  , mouseYInWorld   = screenToWorld(mouseXOnScreen, mouseYOnScreen)

	local text
		=  string.format("filter = %s",            canvasFilter)
		.. string.format("\nsubpixels = %dx%d",    canvasSubpixels, canvasSubpixels)
		.. string.format("\ninteger scaling = %s", integerScaling and "on" or "off")
		.. string.format("\ncanvas scale = %.2f",  canvasScale)
		.. "\n"
		.. string.format("\nplayer pos: world=[%d, %d] screen=[%d, %d]", player.x,player.y, playerXOnScreen,playerYOnScreen)
		.. string.format("\ncamera pos: world=[%d, %d] screen=[%d, %d]", camera.x,camera.y, cameraXOnScreen,cameraYOnScreen)
		.. string.format("\nmouse pos: world=[%d, %d] screen=[%d, %d]",  mouseXInWorld,mouseYInWorld, mouseXOnScreen,mouseYOnScreen)
		.. "\n"
		.. "\n1: no subpixels, nearest filter"
		.. "\n2: no subpixels, linear filter"
		.. "\n3: few subpixels"
		.. "\n4: more subpixels"
		.. "\ni: toggle integer scaling"
		.. "\narrows: move player"
		.. "\n"
		.. "\nTry settings at different window sizes!"
	love.graphics.print(text, 5, 5)
end


